#!/usr/bin/env bats   -*- bats -*-
#
# Tests for podman update
#

load helpers

LOOPDEVICE=

function teardown() {
    if [[ -n "$LOOPDEVICE" ]]; then
        losetup -d $LOOPDEVICE
        LOOPDEVICE=
    fi
    basic_teardown
}


# bats test_tags=distro-integration
@test "podman update - test all options" {
    local cgv=1
    if is_cgroupsv2; then
        cgv=2;
    fi

    # Need a block device for blkio-weight-device testing
    local pass_loop_device=
    if ! is_rootless; then
        if is_cgroupsv2; then
            lofile=${PODMAN_TMPDIR}/disk.img
            fallocate -l 1k  ${lofile}
            LOOPDEVICE=$(losetup --show -f $lofile)
            pass_loop_device="--device $LOOPDEVICE"

            # Get maj:min (tr needed because losetup seems to use %2d)
            lomajmin=$(losetup -l --noheadings --output MAJ:MIN $LOOPDEVICE | tr -d ' ')
        fi
    fi

    # Shortcuts to make the table narrower
    local -a gig=(0 1073741824 2147483648 3221225472)
    local devicemax="1:5 rbps=10485760 wbps=31457280 riops=2000 wiops=4000"
    local mm=memory/memory

    # Format:
    #   --<option> = <value>  | rootless? | check: cgroups v1            | check: cgroups v2
    #
    # Requires very wide window to read. Sorry.
    #
    # FIXMEs:
    #  cpu-rt-period  (cgv1 only, on cpu/cpu.rt_period_us) works on RHEL8 but not on Ubuntu
    #  cpu-rt-runtime (cgv1 only, on cpu/cpu.rt_runtime_us) fails: error setting cgroup config for procHooks ...
    tests="
cpu-shares          =            512 | - | cpu/cpu.shares       = 512              | cpu.weight      = 20
cpus                =              5 | - | cpu/cpu.cfs_quota_us = 500000           | cpu.max         = 500000 100000
cpuset-cpus         =              0 | - | cpuset/cpuset.cpus   = 0                | cpuset.cpus     = 0
cpuset-mems         =              0 | - | cpuset/cpuset.mems   = 0                | cpuset.mems     = 0

memory              =             1G | 2 | $mm.limit_in_bytes       = ${gig[1]}    | memory.max      = ${gig[1]}
memory-swap         =             3G | 2 | $mm.memsw.limit_in_bytes = ${gig[3]}    | memory.swap.max = ${gig[2]}
memory-reservation  =           400M | 2 | $mm.soft_limit_in_bytes  = 419430400    | memory.low      = 419430400

blkio-weight        =            321 | - | -                                       | io.bfq.weight   = default 321 $lomajmin 98
blkio-weight-device = $LOOPDEVICE:98 | - | -                                       | io.bfq.weight   = default 321 $lomajmin 98

device-read-bps     = /dev/zero:10mb | - | -                                       | io.max          = $devicemax
device-read-iops    = /dev/zero:2000 | - | -                                       | io.max          = $devicemax
device-write-bps    = /dev/zero:30mb | - | -                                       | io.max          = $devicemax
device-write-iops   = /dev/zero:4000 | - | -                                       | io.max          = $devicemax
"

    # Run a container
    run_podman run ${pass_loop_device} -d $IMAGE sleep infinity
    cid="$output"

    # Pass 1: read the table above, gather up the options applicable
    # to this test environment (root/rootless, cgroups v1/v2)
    local -a opts
    local -A check
    while read opt works_rootless cgv1 cgv2; do
        if is_rootless; then
            local skipping="skipping --$opt : does not work rootless"
            if [[ $works_rootless = '-' ]]; then
                echo "[ $skipping ]"
                continue
            fi
            if [[ ! $works_rootless =~ $cgv ]]; then
                echo "[ $skipping on cgroups v$cgv ]"
                continue
            fi
        fi

        # Determine the "path = newvalue" string for this cgroup
        tuple=$cgv1
        if is_cgroupsv2; then
            tuple=$cgv2
        fi
        if [[ $tuple = '-' ]]; then
            echo "[ skipping --$opt : N/A on cgroups v$cgv ]"
            continue
        fi

        # Sigh. bfq doesn't exist on Debian (2024-03)
        read path op expect <<<"$tuple"
        if [[ ! -e /sys/fs/cgroup/$path ]]; then
            echo "[ skipping --$opt : /sys/fs/cgroup/$path does not exist ]"
            continue
        fi

        # OK: setting is applicable. Preserve it. (First removing whitespace)
        opt=${opt// /}
        opts+=("--$opt")
        check["--$opt"]=$tuple
    done < <(parse_table "$tests")

    # Now do the update in one fell swoop
    run_podman update "${opts[@]}" $cid

    # ...and check one by one
    defer-assertion-failures
    for opt in "${opts[@]}"; do
        read path op expect <<<"${check[$opt]}"
        run_podman exec $cid cat /sys/fs/cgroup/$path

        # Magic echo of unquoted-output converts newlines to spaces;
        # important for otherwise multiline blkio file.
        updated="$(echo $output)"
        assert "$updated" $op "$expect" "$opt ($path)"
    done
    immediate-assertion-failures

    # Clean up
    run_podman rm -f -t0 $cid
    if [[ -n "$LOOPDEVICE" ]]; then
        losetup -d $LOOPDEVICE
        LOOPDEVICE=
    fi
}

@test "podman update - set restart policy" {
    touch ${PODMAN_TMPDIR}/sentinel
    run_podman run --security-opt label=disable --name testctr -v ${PODMAN_TMPDIR}:/testdir -d $IMAGE sh -c "touch /testdir/alive; while test -e /testdir/sentinel; do sleep 0.1; done;"

    run_podman container inspect testctr --format "{{ .HostConfig.RestartPolicy.Name }}"
    is "$output" "no"

    run_podman update --restart always testctr

    run_podman container inspect testctr --format "{{ .HostConfig.RestartPolicy.Name }}"
    is "$output" "always"

    # Ensure the container is alive
    wait_for_file ${PODMAN_TMPDIR}/alive

    rm -f ${PODMAN_TMPDIR}/alive
    rm -f ${PODMAN_TMPDIR}/sentinel

    # Restart should ensure that the container comes back up and recreates the file
    wait_for_file ${PODMAN_TMPDIR}/alive

    run_podman rm -f -t0 testctr
}

function _create_container_check_change_of_HealthCheck_configuration_with_podman_update(){
    local ctrname="$1"
    local msg="$2"
    local format="$3"
    local flag="$4"
    local value="$5"
    local expect="$6"
    local expect_msg="$7"
    local expect_exit_code="$8"
    local is_startup="$9"
    local no_hc="${10}"

    if [[ $is_startup = "yes" ]]; then
        run_podman run -d --name $ctrname    \
            --health-cmd "echo $msg"         \
            --health-startup-cmd "echo $msg" \
            $IMAGE /home/podman/pause
        cid="$output"
    else
        if [[ $no_hc = "yes" ]]; then
            run_podman run -d --name $ctrname \
                    $IMAGE /home/podman/pause
            cid="$output"
        else
            run_podman run -d --name $ctrname \
                --health-cmd "echo $msg"  \
                $IMAGE /home/podman/pause
            cid="$output"
        fi
    fi


    if [ -n "${value}" ]; then
        run_podman $expect_exit_code update $ctrname $flag "$value"
    else
        run_podman $expect_exit_code update $ctrname $flag
    fi

    if [[ $expect_exit_code = 0 ]]; then
        run_podman inspect $ctrname --format $format
        assert "$output" == "$expect" "$expect_msg"
    fi

    output=$cid
}

# HealthCheck configuration

@test "podman update - --health-cmd" {
    local msg="healthmsg-$(random_string)"
    local new_msg="healthmsg-new-msg"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                          \
        $msg                              \
        "{{.Config.Healthcheck.Test}}"    \
        "--health-cmd" "echo $new_msg"    \
        "[CMD-SHELL echo $new_msg]"       \
        ".Config.Healthcheck.Test"        \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-interval" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                          \
        $msg                              \
        "{{.Config.Healthcheck.Interval}}"\
        "--health-interval" "10s"         \
        "10s"                             \
        ".Config.Healthcheck.Interval"    \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-log-destination" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"
    local TMP_DIR_HEALTHCHECK="$PODMAN_TMPDIR/healthcheck"
    mkdir $TMP_DIR_HEALTHCHECK

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                           \
        $msg                                               \
        "{{.Config.HealthLogDestination}}"                 \
        "--health-log-destination" "$TMP_DIR_HEALTHCHECK"  \
        "$TMP_DIR_HEALTHCHECK"                             \
        ".Config.HealthLogDestination"                     \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-max-log-count" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                          \
        $msg                              \
        "{{.Config.HealthMaxLogCount}}"   \
        "--health-max-log-count" "20"     \
        "20"                              \
        ".Config.HealthMaxLogCount"       \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-max-log-size" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                          \
        $msg                              \
        "{{.Config.HealthMaxLogSize}}"    \
        "--health-max-log-size" "10"      \
        "10"                              \
        ".Config.HealthMaxLogSize"        \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-on-failure" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                             \
        $msg                                                 \
        "{{.Config.HealthcheckOnFailureAction}}" \
        "--health-on-failure" "restart"                      \
        "restart"                                            \
        ".Config.Healthcheck.HealthcheckOnFailureAction"     \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-retries" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                          \
        $msg                              \
        "{{.Config.Healthcheck.Retries}}" \
        "--health-retries" "5"            \
        "5"                               \
        ".Config.Healthcheck.Retries"     \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-timeout" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                          \
        $msg                              \
        "{{.Config.Healthcheck.Timeout}}" \
        "--health-timeout" "10s"          \
        "10s"                             \
        ".Config.Healthcheck.Timeout"     \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --no-healthcheck" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                          \
        $msg                              \
        "{{.Config.Healthcheck.Test}}"    \
        "--no-healthcheck" ""             \
        "[NONE]"                          \
        "HealthCheck command is disabled" \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}


@test "podman update - --health-start-period" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                              \
        $msg                                  \
        "{{.Config.Healthcheck.StartPeriod}}" \
        "--health-start-period" "10s"         \
        "10s"                                 \
        ".Config.Healthcheck.StartPeriod"     \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - no hc command --health-interval" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                          \
        $msg                              \
        "{{.Config.Healthcheck.Interval}}"\
        "--health-interval" "10s"         \
        "10s"                             \
        ".Config.Healthcheck.Interval"    \
        125 "no" "yes"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - no hc command --health-retries" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                          \
        $msg                              \
        "{{.Config.Healthcheck.Retries}}" \
        "--health-retries" "5"            \
        "5"                               \
        ".Config.Healthcheck.Retries"     \
        125 "no" "yes"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - no hc command --health-timeout" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                          \
        $msg                              \
        "{{.Config.Healthcheck.Timeout}}" \
        "--health-timeout" "10s"          \
        "10s"                             \
        ".Config.Healthcheck.Timeout"     \
        125 "no" "yes"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - no hc command --health-start-period" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                              \
        $msg                                  \
        "{{.Config.Healthcheck.StartPeriod}}" \
        "--health-start-period" "10s"         \
        "10s"                                 \
        ".Config.Healthcheck.StartPeriod"     \
        125 "no" "yes"

    run_podman rm -t 0 -f $ctrname
}

# Startup HealthCheck Configuration

@test "podman update - --health-startup-cmd" {
    local msg="healthmsg-$(random_string)"
    local new_msg="healthmsg-new-msg"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                 \
        $msg                                     \
        "{{.Config.StartupHealthCheck.Test}}"    \
        "--health-startup-cmd" "echo $new_msg"   \
        "[CMD-SHELL echo $new_msg]"              \
        ".Config.StartupHealthCheck.Test"        \
        0 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-startup-interval" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                  \
        $msg                                      \
        "{{.Config.StartupHealthCheck.Interval}}" \
        "--health-startup-interval" "10s"         \
        "10s"                                     \
        ".Config.StartupHealthCheck.Interval"     \
        0 "yes"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-startup-retries" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                 \
        $msg                                     \
        "{{.Config.StartupHealthCheck.Retries}}" \
        "--health-startup-retries" "5"           \
        "5"                                      \
        ".Config.StartupHealthCheck.Retries"     \
        0 "yes"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-startup-success" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                   \
        $msg                                       \
        "{{.Config.StartupHealthCheck.Successes}}" \
        "--health-startup-success " "10"           \
        "10"                                       \
        ".Config.StartupHealthCheck.Successes"     \
        0 "yes"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - --health-startup-timeout" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                 \
        $msg                                     \
        "{{.Config.StartupHealthCheck.Timeout}}" \
        "--health-startup-timeout" "10s"         \
        "10s"                                    \
        ".Config.StartupHealthCheck.Timeout"     \
        0 "yes"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - no startup hc command --health-startup-interval" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                  \
        $msg                                      \
        "{{.Config.StartupHealthCheck.Interval}}" \
        "--health-startup-interval" "10s"         \
        "10s"                                     \
        ".Config.StartupHealthCheck.Interval"     \
        125 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - no startup hc command --health-startup-retries" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                 \
        $msg                                     \
        "{{.Config.StartupHealthCheck.Retries}}" \
        "--health-startup-retries" "5"           \
        "5"                                      \
        ".Config.StartupHealthCheck.Retries"     \
        125 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - no startup hc command --health-startup-success" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                   \
        $msg                                       \
        "{{.Config.StartupHealthCheck.Successes}}" \
        "--health-startup-success" "10"            \
        "10"                                       \
        ".Config.StartupHealthCheck.Successes"     \
        125 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - no startup hc command --health-startup-timeout" {
    local msg="healthmsg-$(random_string)"
    local ctrname="c-h-$(safename)"

    _create_container_check_change_of_HealthCheck_configuration_with_podman_update \
        $ctrname                                 \
        $msg                                     \
        "{{.Config.StartupHealthCheck.Timeout}}" \
        "--health-startup-timeout" "10s"         \
        "10s"                                    \
        ".Config.StartupHealthCheck.Timeout"     \
        125 "no"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - check behavior - change cmd and destination healthcheck" {
    local TMP_DIR_HEALTHCHECK="$PODMAN_TMPDIR/healthcheck"
    mkdir $TMP_DIR_HEALTHCHECK
    local ctrname="c-h-$(safename)"
    local msg="healthmsg-$(random_string)"

    run_podman run -d --name $ctrname     \
                --health-cmd "echo $msg"  \
                $IMAGE /home/podman/pause
    cid="$output"

    run_podman healthcheck run $ctrname
    is "$output" "" "output from 'podman healthcheck run'"

    run_podman update $ctrname --health-cmd "echo healthmsg-new" --health-log-destination $TMP_DIR_HEALTHCHECK

    run_podman healthcheck run $ctrname
    is "$output" "" "output from 'podman healthcheck run'"

    healthcheck_log_path="${TMP_DIR_HEALTHCHECK}/${cid}-healthcheck.log"
    # The healthcheck is triggered by the podman when the container is started, but its execution depends on systemd.
    # And since `run_podman healthcheck run` is also run manually, it will result in two runs.
    count=$(grep -co "healthmsg-new" $healthcheck_log_path)
    assert "$count" -ge 1 "Number of matching health log messages"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - check behavior - change healthcheck interval" {
    local ctrname="c-h-$(safename)"
    local msg="healthmsg-$(random_string)"

    run_podman run -d --name $ctrname     \
                --health-cmd "echo $msg"  \
                --health-interval 2s      \
                $IMAGE top
    cid="$output"

    sleep 2s

    run_podman update $ctrname --health-interval 1s

    sleep 5s

    run_podman inspect $ctrname --format "{{.State.Health.Log}}"
    count=$(grep -co "$msg" <<< "$output")
    assert "$count" -ge 3 "Number of matching health log messages"

    run_podman rm -t 0 -f $ctrname
}


@test "podman update - check behavior - set healthcheck interval" {
    local ctrname="c-h-$(safename)"
    local msg="healthmsg-$(random_string)"

    run_podman run -d --name $ctrname     \
                --health-cmd "echo $msg"  \
                --health-interval disable \
                $IMAGE top
    cid="$output"

    run_podman update $ctrname --health-interval 1s

    sleep 5s

    run_podman inspect $ctrname --format "{{.State.Health.Log}}"
    count=$(grep -co "$msg" <<< "$output")
    assert "$count" -ge 3 "Number of matching health log messages"

    run_podman rm -t 0 -f $ctrname
}

@test "podman update - check behavior - change healthcheck startup interval" {
    local ctrname="c-h-$(safename)"
    local msg="healthmsg-$(random_string)"

    run_podman run -d --name $ctrname                   \
                --health-cmd "echo normal$msg"          \
                --health-startup-cmd "echo startup$msg" \
                --health-startup-interval 30s           \
                --health-startup-success 3              \
                $IMAGE top
    cid="$output"

    sleep 1s

    run_podman update $ctrname --health-startup-interval 1s

    sleep 5s

    run_podman inspect $ctrname --format "{{.State.Health.Log}}"
    count=$(grep -co "startup$msg" <<< "$output")
    assert "$count" -ge 3 "Number of matching startup health log messages"

    count=$(grep -co "normal$msg" <<< "$output")
    assert "$count" -ge 1 "Number of matching health log messages"

    run_podman rm -t 0 -f $ctrname
}


@test "podman update - check behavior - set healthcheck startup interval" {
    local ctrname="c-h-$(safename)"
    local msg="healthmsg-$(random_string)"

    run_podman run -d --name $ctrname                   \
                --health-cmd "echo normal$msg"          \
                --health-startup-cmd "echo startup$msg" \
                --health-startup-interval disable       \
                $IMAGE top
    cid="$output"

    sleep 1s

    run_podman update $ctrname --health-startup-interval 1s

    sleep 5s

    run_podman inspect $ctrname --format "{{.State.Health.Log}}"
    count=$(grep -co "$msg" <<< "$output")
    assert "$count" -ge 1 "Number of matching health log messages"

    run_podman rm -t 0 -f $ctrname
}
# vim: filetype=sh
